.sig + .sig {
  border-top-left-radius: 0;
  border-top-right-radius: 0;
  padding-top: .05rem;
}
.sig:has(+.sig) {
  border-bottom-left-radius: 0;
  border-bottom-right-radius: 0;
  padding-bottom: .05rem;
}
